import gcsfs
import datetime

def get_last_date(json_gs):
        files = gcsfs.GCSFileSystem().ls(json_gs, encoding='utf-8')
        dates = []
        for file in files:
            date = file.split("/")[-1]
            if date != '':
                dates.append(date)
            
        # print(dates)
        date_objects = [datetime.datetime.strptime(date_str, '%Y-%m-%d').date() for date_str in dates]
        max_date = max(date_objects)
        return max_date


from google.cloud import bigquery
from google.cloud.exceptions import NotFound
from data_code.utils.appearence import highlight_text
from data_code.utils.evaluate import evaluate_response_bq
import subprocess
import json
import ast

def create_or_evaluate_bq_table(table_id:str,trn_parquet:str, delete_columns:str)->dict:
        """Verifica si existe la tabla y extrae los metadatos; en caso de que no exista la crea con base en el parquet de raw.

        Args:
            table (str): Nombre de la tabla.
            raw_parquet (str): Ruta del parquet.

        Returns:
            dict: Diccionario con la metadata de la tabla.
        """
        try:
            client = bigquery.Client()
            client.get_table(table_id.replace(':','.'))  # Make an API request.
            print(f"Table: {highlight_text(table_id, 32)} already exists.")
            metadata_table_bq = get_metadata(table_id)['schema']['fields']
            
            # validate_columns_hom(table_id)

            return metadata_table_bq
        except NotFound:
            print(f"Table: {highlight_text(table_id,31)} not found. \nCreating it with parquet: {highlight_text(trn_parquet,34)}")
            response = subprocess.run(["bq","load","--autodetect", "--source_format=PARQUET", table_id, trn_parquet], stdout=subprocess.PIPE, stderr=subprocess.PIPE, shell=True)
            message = evaluate_response_bq(response,table_id)
            print(message)
            metadata_table_bq = get_metadata(table_id)
            if delete_columns != "[]":
                print("Erasing columns")
                delete_columns_processed = ast.literal_eval(delete_columns.replace('"',''))   
                delete_columns_table(table_id,delete_columns_processed)
                print(f"Columns: {delete_columns} erased.")
                # Procesamiento de hom
                print("\Processing hom")
                validate_columns_hom(table_id)
            return metadata_table_bq

def validate_columns_hom(table_id):
    table_id_hom = table_id.replace('raw','hom').replace('wrk','hom')
    print(f"""
        Comparing Beetween:
          Wrk: {highlight_text(table_id,33)}
          Hom: {highlight_text(table_id_hom,33)}      
        """)
    metadata_table_bq = get_metadata(table_id)['schema']['fields']
    wrk_json= json.loads(json.dumps(metadata_table_bq,sort_keys=True))
    metadata_table_bq_hom = get_metadata(table_id_hom)['schema']['fields']
    hom_json = json.loads(json.dumps(metadata_table_bq_hom,sort_keys=True))
    for wrk_data in wrk_json:
        wrk_name = wrk_data['name']
        wrk_type = wrk_data['type']                    
        print(wrk_name)
        for hom_data in hom_json:
            hom_name = hom_data['name']
            hom_type = hom_data['type']
            if wrk_name == hom_name and wrk_type != hom_type:
                print(f"Work: {wrk_data}\nHom: {hom_data}")
                fix_column_data_type(table_id, wrk_name, hom_data)
            if wrk_name == hom_name and wrk_type == hom_type:
                print(highlight_text("Nothing to change",34))
        print("------------------")
        
def delete_columns_table(table_id:str,delete_columns:str):
    client = bigquery.Client()

    table_id = table_id.replace(":",".")
    for column in delete_columns:
        delete_columns_query = f'''
                                ALTER TABLE {table_id}
                                DROP COLUMN {column}
                                '''
        print(delete_columns_query)
        query_job = client.query(delete_columns_query)
        query_job.result()

def get_metadata(table_id:str)->dict:
    """_summary_

    Args:
        table_id (str): _description_

    Returns:
        dict: _description_
    """
    metadata_table_bq = subprocess.run(["bq","show","--format=prettyjson", table_id], capture_output=True, shell=True).stdout
    metadata_table_bq = ast.literal_eval(metadata_table_bq.decode('UTF-8'))
    return metadata_table_bq

def fix_column_data_type(table_id:str, data_wrk_name:str, data_hom:dict):
    client = bigquery.Client()
    print(f"Fixing field {data_wrk_name}")
    
    table_id = table_id.replace(":",".")
    dropping_query = f"""
                ALTER TABLE {table_id} 
                DROP COLUMN {data_wrk_name};
                """
    print(dropping_query)
    query_job = client.query(dropping_query)
    query_job.result()
    data_type = convert_data_types(data_hom['type'])
    creating_query = f"""
                    ALTER TABLE {table_id} 
                    ADD COLUMN {data_wrk_name} {data_type}
                    """
    if data_hom['description']:
        creating_query = f"{creating_query} OPTIONS (description='{data_hom['description']}');"
    else:
        creating_query += ";"
    
    print(creating_query)
    query_job = client.query(creating_query)
    query_job.result()

def convert_data_types(to_convert:str):
    types = {
        "ARRAY":"ARRAY",
        "BIGNUMERIC":"BIGNUMERIC",
        "BOOL":"BOOL",
        "BYTES":"BYTES",
        "DATE":"DATE",
        "DATETIME":"DATETIME",
        "FLOAT":"FLOAT64",
        "GEOGRAPHY":"GEOGRAPHY",
        "INTEGER":"INT64",
        "INTERVAL":"INTERVAL",
        "JSON":"JSON",
        "NUMERIC":"NUMERIC",
        "STRING":"STRING",
        "STRUCT":"STRUCT",
        "TIME":"TIME",
        "TIMESTAMP":"TIMESTAMP",
    }
    return types[to_convert]



import gcsfs
import json

def get_conf_environment(env:str)->tuple:
    """_summary_

    Args:
        env (str): _description_

    Raises:
        KeyError: _description_

    Returns:
        str: _description_
    """

    if env == "dev":
        uri_base = "gs://us-east1-yas-dev-composer-e-edde754f-bucket/data/yas-dwh"
        conf_file = f"{uri_base}/conf.json"
        connections = f"{uri_base}/db_connections.json"

    elif env == "qas":
        uri_base = "gs://us-east1-yas-qas-composer-e-40e740d3-bucket/data/yas-dwh"
        conf_file = f"{uri_base}/conf.json"
        connections = f"{uri_base}/db_connections.json"

    elif env == "int":
        uri_base = "gs://us-east1-yas-int-composer-e-83ac79a1-bucket/data/yas-dwh"
        conf_file = f"{uri_base}/conf.json"
        connections = f"{uri_base}/db_connections.json"

    elif env == "prd":
        uri_base = "gs://us-east1-yas-prd-composer-e-b710217f-bucket/data/yas-dwh"
        conf_file = f"{uri_base}/conf.json"
        connections = f"{uri_base}/db_connections.json"
    else:
        raise KeyError(f"Invalid environment {env}")
    # print("conf: ",conf_file)
    return conf_file,connections

def evaluate_app(app:str)->str:
    """Take in consideration if is necesary to add more information in the app chain.

    Args:
        app (str): Which app es going to work

    Returns:
        str: Chain nedeed in order to access to the elements. (templates, staging, etc.)
    """
    #TODO: If app still have cuentas...
    return app

def get_params(params_file:str, environment:str)->dict:
    """Get information related to each process in the paramaters file.

    Args:
        params_file (str): Uri file with the parameters in JSON format.
        environment (str): Which environment is going to run.

    Returns:
        dict: The whole information needed in order to process dataflow.
    """
    print(f"params:\t{params_file}\nenv:\t{environment}")
    with gcsfs.GCSFileSystem().open(params_file,encoding='utf-8') as parameters:
        jd = json.load(parameters)
        conf_file, connections = get_conf_environment(environment)
        with gcsfs.GCSFileSystem().open(conf_file,encoding='utf-8') as config:
            conf = json.load(config)

            # Datos generales
            service_account_email = conf["service_account_dwh"]
            setup_file = conf["setup_file"]
            runner = conf["runner"]
            region = conf["region"]
            project = conf["proy_dwh"]
            dataset_bq= conf["dataset_l"]

            app = jd['app']
            template_name = jd["template_name"]
            code_file = jd["code_file"]
            config_name = jd["config_file"]


            app = evaluate_app(app)
            uri = f"gs://yas-{environment}-dwh-des/{app}"
            template_location = f"{uri}/tpl/{template_name}"
            config_file = f"{uri}/param/{config_name}"
            staging_location = f"gs://yas-{environment}-dwh-staging/"
            temp_location = f"gs://yas-{environment}-dwh-tmp/"
            print(f'uri:\t{uri}')
                    

            params = {
                "setup_file":setup_file,
                "service_account_email":service_account_email,
                "project":project,
                "runner":runner,
                "region":region,
                "template_location":template_location,
                "config_file":config_file,
                "staging_location":staging_location,
                "temp_location":temp_location,
                "code_file":code_file,
                "dataset_bq":dataset_bq
            }

            if code_file=='./DFC_DWH_BCH_CAN_SQL.py':
                json_file = open("connections/app_database.json", "r").read()
                app_dict = json.loads(json_file)
                database = app_dict[jd['app']]
                print(f"database:\t{database}")
                database_connections = json.load(gcsfs.GCSFileSystem().open(connections,encoding='utf-8'))
                creds = database_connections[database]
                user = creds['user']
                pss = creds['pss']
                psswd = creds['psswd']
                db = creds['dbname']
                host = creds['host']
                port = creds['port']
                database_type = creds['type']
                certs_location = f"{uri}/security/"

                # General
                
                gen = database_connections["General"]
                sslmode = gen["sslmode"]
                sslrootcert = gen["sslrootcert"]
                sslcert = gen["sslcert"]
                sslkey = gen["sslkey"]
                schema = gen["schema"]

                sql_params = {
                    "user":user,
                    "pss":pss,
                    "psswd":psswd,
                    "db":db,
                    "host":host,
                    "port":port,
                    "type":database_type,
                    "certs_location":certs_location,
                    "sslmode": sslmode,
                    "sslrootcert": sslrootcert,
                    "sslcert": sslcert,
                    "sslkey": sslkey,
                    "schema": schema,
                    "app":jd['app']
                }

                params.update(sql_params)

                print(params)
                

    return params


import apache_beam as beam

# SQL Utility 

def format_dict(element):
    formato = {
        "column_name":element[0],
        "udt_name":element[1]
    }
    return formato

class FormatRecords(beam.DoFn):
    def process(self, elements):
        for record in elements[1]:
            formato = {
            "column_name":record[0],
            "udt_name":record[1]
            }
            yield formato

# RAW Utility

class SeparateRecords(beam.DoFn):
    def process(self, elements):
        for element in elements:
            yield element

def concat_headers(element):
    info=[]
    names= []
    record = {}
    for name in element[0]:
        names.append(name.name)
    for row in element[1]:
        record = {}
        for name,data in zip(names,row):
            record[name] = data
        info.append(record)
    return info

# TRN Utility

# WRK Utility